/***************************************************************************
                          prototypedoc.cpp  -  description
                             -------------------
    begin                : Thu May 15 22:50:15 CEST 2003
    copyright            : (C) 2003 by Marcus Beranek
    email                : marcus@beranek.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// This class handles the document-parsing of the XML-file. I haven't
// worked with QTs XML-functions, so I've used the libxml2 for parsing
// the XML-files.
//
// ProtoTypeDoc also constructs the adjacency-list of the blocks (see "block.cpp").
//

#include "prototypedoc.h"

#include "block.h"
#include "blocklist.h"
#include "successor.h"


#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <iostream>
#include <iomanip>
#include <qptrlist.h>
#include <qdir.h>

ProtoTypeDoc::ProtoTypeDoc()
{
	modified = false;
	blocklist = new BlockList();
	maxBlockID = 0;
	currentProtocol = "";
	currentProtocolFilename = "";
}
//---------------------------------------------------------------------------


ProtoTypeDoc::~ProtoTypeDoc()
{
}
//---------------------------------------------------------------------------


void ProtoTypeDoc::newDoc()
{
}
//---------------------------------------------------------------------------

//
// Just a "read-only"-tool for now. Maybe later there will be the possibility
// to store the documentation of your own protocol ...
//
bool ProtoTypeDoc::save()
{
	return true;
}
//---------------------------------------------------------------------------


bool ProtoTypeDoc::saveAs(const QString &filename)
{
	return true;
}
//---------------------------------------------------------------------------


bool ProtoTypeDoc::load(const QString &filename)
{

	// Clear the blocklist, insert root node and parse XML-document ...
	blocklist->clear();
	startblock = new Block();
	startblock->ID=0;
	startblock->Depth=0;
	startblock->Name="Root-Node";
	startblock->visited=false;
	blocklist->append( startblock );
	maxBlockID = 0;    // only startblock-ID

	xmlDocPtr document;
	xmlNodePtr cur;
	
	// TODO: check if signedness is important with filenames...
	char* docname = filename.latin1();
	document = xmlParseFile(docname);

	
	x_coord = 0;

	// Ab hier im wesentlichen aus der libxml2-Doku bernommen...
	if (document == NULL ) {
		fprintf(stderr,"Document not parsed successfully. \n");
	} else {

		cur = xmlDocGetRootElement(document);

		if (cur == NULL) {
			fprintf(stderr,"empty document\n");
			xmlFreeDoc(document);
		} else {

			if (xmlStrcmp(cur->name, (const xmlChar *) "protocol")) {
				fprintf(stderr,"document of the wrong type, root node != protocol");
				xmlFreeDoc(document);
			} else {

				cur = cur->xmlChildrenNode;
				while (cur != NULL) {
					if ((!xmlStrcmp(cur->name, (const xmlChar *)"protocolname"))){

						xmlChar *key = xmlNodeListGetString(document, cur->xmlChildrenNode, 1);
						// TODO: Check if this conversion is correct!!!
						//       libxml2 uses "unsigned char", not "char"
						QString p_name = (char*) key;
						startblock->Name = p_name.simplifyWhiteSpace();;
						startblock->Command = p_name.simplifyWhiteSpace();;
						currentProtocol = p_name;

					}
					if ((!xmlStrcmp(cur->name, (const xmlChar *)"protocoldescription"))){

						xmlChar *key = xmlNodeListGetString(document, cur->xmlChildrenNode, 1);
						// TODO: Check if this conversion is correct!!!
						//       libxml2 uses "unsigned char", not "char"
						QString p_desc = (char*) key;
						startblock->Description = p_desc.simplifyWhiteSpace();;

					}

					if ((!xmlStrcmp(cur->name, (const xmlChar *)"protocolopening"))){

						xmlChar *key = xmlNodeListGetString(document, cur->xmlChildrenNode, 1);
						// TODO: Check if this conversion is correct!!!
						//       libxml2 uses "unsigned char", not "char"
						QString p_open = (char*) key;
						int po = p_open.simplifyWhiteSpace().toInt();
						if ( po > 0 ){
							sux = new Successor();
							Successor &s = *sux;
							s.ID = po;
							startblock->addTypicalFollowedby( s );
						}

					}

					if ((!xmlStrcmp(cur->name, (const xmlChar *)"block"))){

						Block* block = new Block( );

						xmlNodePtr cur2 = cur->xmlChildrenNode;
						xmlChar* key = xmlGetProp(cur, "id");

						// TODO: Check if this conversion is correct!!!
						//       libxml2 uses "unsigned char", not "char"
						QString blockid = (char*) key;
						int bid = blockid.simplifyWhiteSpace().toInt();

						// remember maximum Block-ID
						maxBlockID = ( maxBlockID > bid ) ? ( maxBlockID ) : ( bid );

						key = xmlGetProp(cur, "type");

						// TODO: Check if this conversion is correct!!!
						//       libxml2 uses "unsigned char", not "char"
						QString blocktype = (char*) key;

						if ( blocktype.lower().compare( QString( "send" ) ) == 0 ){
							block->Type = Block::Send;
						}
						if ( blocktype.lower().compare( QString( "receive" ) ) == 0 ){
							block->Type = Block::Receive;
						}

						block->ID = bid;

						while(cur2 != NULL){

							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockname"))){

								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockname = (char*) key;
								block->Name = blockname.simplifyWhiteSpace();

							}

							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockcommand"))){

								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockcmd = (char*) key;
								block->Command = blockcmd.simplifyWhiteSpace();

							}


							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockfunction"))){

								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockfunc = (char*) key;
								block->Function = blockfunc.simplifyWhiteSpace();

							}

							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockstate"))){

								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockstate = (char*) key;
								block->State = blockstate.simplifyWhiteSpace();

							}

							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockexample"))){

								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockexample = (char*) key;
								block->Example = stripBlanks(blockexample);

							}

							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockdescription"))){


								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								//                AnsiString keyAnsi = (char*) key;
								QString blockdesc = (char*) key;
								block->Description = stripBlanks(blockdesc);

							}


							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockfollows"))){


								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockfls = (char*) key;
								int bf = blockfls.simplifyWhiteSpace().toInt();
								if ( bf == 0 ){
									sux = new Successor();
									Successor &s = *sux;
									s.ID = block->ID;
									startblock->addFollowedby( s );
								} else {
									// TODO: fill Follows-list fls
									sux = new Successor();
									Successor &s = *sux;
									s.ID = bf;
									block->addFollows( s );
								}
								// hier mu man ein bisschen aufpassen, da ja noch nicht alle
								// Komponenten der XML-Datei eingelesen sind und somit noch nicht alle
								// Blocks in der Liste existieren, deswegen auch der Hilfstyp Successor...

							}


							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"typicalfollowedby"))){


								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blocktfby = (char*) key;
								int bfb = blocktfby.simplifyWhiteSpace().toInt();
								if ( bfb > 0 ){
									sux = new Successor();
									Successor &s = *sux;
									s.ID = bfb;
									block->addTypicalFollowedby( s );
								}
								// hier mu man ein bisschen aufpassen, da ja noch nicht alle
								// Komponenten der XML-Datei eingelesen sind und somit noch nicht alle
								// Blocks in der Liste existieren, deswegen auch der Hilfstyp Successor...

							}


							if ((!xmlStrcmp(cur2->name, (const xmlChar *)"blockfollowedby"))){


								xmlChar *key = xmlNodeListGetString(document, cur2->xmlChildrenNode, 1);

								// TODO: Check if this conversion is correct!!!
								//       libxml2 uses "unsigned char", not "char"
								QString blockflb = (char*) key;
								int bfb = blockflb.simplifyWhiteSpace().toInt();
								if ( bfb > 0 ){
									sux = new Successor();
									Successor &s = *sux;
									s.ID = bfb;
									block->addFollowedby( s );
								}
								// hier mu man ein bisschen aufpassen, da ja noch nicht alle
								// Komponenten der XML-Datei eingelesen sind und somit noch nicht alle
								// Blocks in der Liste existieren, deswegen auch der Hilfstyp TSuccessor...

							}


							// zum nchsten XML-Knoten im Bereich "block" gehen:

							cur2 = cur2->next;

						}   // Ende while

						//
						// We use insertion-sort instead of just appending the blocks:
						//
						// blocklist->append( block );
						blocklist->inSort( block );


					} // Ende if block


					// zum nchsten XML-Knoten im Bereich "protocol" gehen:
					cur = cur->next;

				}

				xmlFreeDoc(document);
				// We're using inSort, so the list is already sorted
				// blocklist->sort();
				rebuildProtocolTree( startblock, 0 );
				clearVisited();
				
				// for debugging only:
				//checkProtocolTree( startblock );
				//clearVisited();
			}
		}
	}
	emit documentChanged();
	return true;
}
//---------------------------------------------------------------------------


// Print protocoll-tree to stdout
// for debugging only ...
void ProtoTypeDoc::checkProtocolTree( Block* b )
{
	Block* block;
	int check;

	// Search for the block b and set the correct depth ...
	check = blocklist->find( b );

	if ( check > -1 ){
		block = blocklist->at( check );
		block->visited=true;

		// print out checked block:
		std::cout << "----------------------------------------------" << std::endl;
		std::cout << "Block-ID: " << block->ID << std::endl;
		std::cout << "Block-Name: " << block->Name << std::endl;
		std::cout << "Block-Depth: " << block->Depth << std::endl;

		// ... then get list of all following blocks ...
		SuccessorList bl = block->getFollowedby();
		Successor* f;

		// ... and for evey successor ...
		for( f=bl.first(); f; f=bl.next() ){

			std::cout << "Block followed by " << f->ID << std::endl;

			int z = blocklist->find( new Block( f->ID ) );
			if ( z > -1 ){
				Block* v = blocklist->at( z );
				if( ! v->visited ){
					// ... decent recursive into subtree ...
					//std::cout << "[Decending into subtree of Block " << block->ID << "] " <<  std::endl;
					//checkProtocolTree( new Block( f->ID ) );
					//std::cout << "[End Decent of Block " << block->ID << "]" <<  std::endl;
				} else {
					std::cout << "[Not decending into subtree of Block " << block->ID << "]" <<  std::endl;
				}
			}

			std::cout << "----------------------------------------------" << std::endl;
			std::cout << std::endl;

		}
	}
}
//---------------------------------------------------------------------------



void ProtoTypeDoc::rebuildProtocolTree( Block* b, int depth )
{
	// The main-purpose of this method was to figure out, how deep the block is stored in
	// the protocol-tree.
	//
	// Originally I had a list, which was organized this way: list-index = block->ID
	// So a block for example with the ID 12 could be accessed via blocklist->at[12].
	// Unfortunatly the QPtrList misses the following features:
	//
	//  1) Stored items must not be null (this may happen if the IDs in the XML-file
	//     are not continuous or are incomplete)
	//
	//  2) The QPtrList doesn't expand automagically, when inserting items beyond the
	//     bounds, like the TList does in Kylix.
	//
	// So the TList in Kylix could be easily traversed like a tree, using a recursive
	// call, starting wit rebuildProtocolTree(0,0), reading all subnodes x from root-node 0,
	// and increasing the level l by calling rebuildProtocolTree( x, l++ )...
	//
	// As a workaround, we have to use the find-method of the QPtrList, which calls
	// the compareItems method for every item in the list! This is not very effective for
	// larger lists... :(
	//
	//
	// The cleaner implementation would be the following:
	//
	//  1) read XML-node (block-ID, Follower and FollowedBys)
	//  2) check if all nodes exist for block-ID, Followers/FollowedBys and
	//     create all nodes if neccessary
	//  3) insert nodes in the tree if they don't exist or
	//     update the nodes if they exist.
	//
	//  ... but ...
	//
	//     ... not enough time ... :(
	//
	//
	// So we just calculate depth of every block in the following way:
	//

	Block* m;
	int modify;

	// Search for the block b and set the correct depth ...
	modify = blocklist->find( b );

	if ( modify > -1 ){
		m = blocklist->at( modify );
		m->Depth = depth;
		m->visited=true;

		// ... then get list of all following blocks ...
		SuccessorList bl = m->getFollowedby();
		Successor* f;

		// ... and for evey successor ...
		for( f=bl.first(); f; f=bl.next() ){

			// ... which we haven't visited yet ...
			int u = blocklist->find( new Block( f->ID ) );
			if ( u > -1 ){
				Block* v = blocklist->at( u );
				if( ! v->visited ){
					// ... increase the depth by one ...
					depth++;
					// .. decent recursive into subtree ...
					rebuildProtocolTree( new Block( f->ID ), depth );
					// ... come back again and decrease the depth-value.
					depth--;
				}
			}
		}
	}
}
//---------------------------------------------------------------------------


void ProtoTypeDoc::clearVisited()
{
	Block* v;
	for( v=blocklist->first(); v; v=blocklist->next() ){
		v->visited = false;
	}
}
//---------------------------------------------------------------------------


void ProtoTypeDoc::clearPrinted()
{
	Block* v;
	for( v=blocklist->first(); v; v=blocklist->next() ){
		v->printed = false;
	}
}
//---------------------------------------------------------------------------


bool ProtoTypeDoc::isModified() const
{
	return modified;
}
//---------------------------------------------------------------------------


void ProtoTypeDoc::setConfigPrototypePath( QString dir )
{
	config_protocol_path = dir;
}
//---------------------------------------------------------------------------

void ProtoTypeDoc::setConfigLanguage( QString lang )
{
	config_language = lang;
}


void ProtoTypeDoc::loadProtocol( QString protocol )
{
	QString f = config_protocol_path;
	f.append( "/" );
	f.append( config_language );
	f.append( "/" );
	f.append( protocol );

	currentProtocolFilename = f;
	
	load( f );

}
//---------------------------------------------------------------------------

BlockList* ProtoTypeDoc::getBlockList( )
{
	return blocklist;
}
//---------------------------------------------------------------------------


QString ProtoTypeDoc::getCurrentProtocol()
{
	return currentProtocol;
}
//---------------------------------------------------------------------------


QString ProtoTypeDoc::stripBlanks( QString txt )
{
	//
	// QString has the methods simplifyWhiteSpace and stripWhiteSpace, which affect _all_
	// whitespace-characters, like blanks, tabs and especially newlines!
	// Since it is common usage in XML to format the file with blanks and tabs, we are
	// going to strip them away.
	//
	// But: we are stripping away tabs and blanks!
	// We still leave the newlines, which we use as indicator for paragraphs.
	//
	// Since we have implemented a printing-method which operates on lines, we are going to
	// insert a linebreak every xxx characters, so we can avoid too long lines in the
	// printing-output.
	//
	// Okay: it would be cleaner to use things like docbook or html for that stuff...
	// One possibility would be the following:
	// Store node-documentation in a seperate file and additionally store that filename
	// in the XML-documentation.
	//
	// Maybe something like that:
	//
	// documentation for SMTP stored in SMTP.xml,
	// documentation for "HELO" stored in smtp_1.html
	// ...
	// <block id='1'>
	//  <blockname>HELO</blockname>
	//  <blockdoc>smtp_1.html</blockdoc>
	//  ...
	// </block>
	//

	
	int max = config_max_characters_per_line;
	int len = txt.length();

	QChar c = QChar(' ');
	QChar old_c = QChar(' ');
	QString txt_new  = "";
	QString txt_tmp = "";

	int i = 0;                    // position in current text
	int last_blank = 0;           // last position of blank in current line
	int line_pos = 0;             // position in current line

	while( i < len ){

		c = txt.at(i);

		// tab = blank
		if ( c == QChar('\t') ){
			c = QChar( ' ' );
		}

		// Character is normal letter (i.e. no blank, no linefeed, no carriage-return):
		if( ( c != QChar(' ') ) && ( c != QChar('\n') )  && ( c != QChar('\r') )  ){
			txt_tmp.append( c.latin1() );
		}

		// Character is blank
		if(  c == QChar(' ')  ){
			// flush tmp-buffer:
			txt_new.append( txt_tmp );
			txt_tmp = "";
			// print blank, only if last character was no blank or linefeed:
			if( (c != old_c) && (old_c != QChar('\n')) && (old_c != QChar('\r')) ){
				txt_tmp.append( c.latin1() );
			}
			last_blank = line_pos;
		}

		// Character is linefeed or carriage-return
		if( (c == QChar('\n')) ||  (c == QChar('\r')) ){
			// tmp-buffer flushen:
			txt_new.append( txt_tmp );
			txt_tmp = "\n";
			line_pos = 0;
		}

		// maximum characters per line reached
		if( line_pos == max){
			// insert line-feed
			if( last_blank > 0){
				// if we already have a blank in this line, put linefeed behind this blank
				txt_new.append("\n");
				// flush tmp-buffer:
				txt_new.append( txt_tmp );
				line_pos = txt_tmp.length();
				txt_tmp = "";
			} else {
				// if we don't have a blank, we have to do a hard break in the word
				txt_tmp.append("\n");
				txt_new.append( txt_tmp );
				line_pos = 0;
				txt_tmp = "";
			}
		}

		// proceed...
		i++;
		line_pos++;
		old_c = QChar( c );

	}

	// if we still have characters in our tmp-buffer, flush it:
	if( ! txt_tmp.isEmpty() ) { txt_new.append( txt_tmp ); }

	// sometimes we have some blanks at the beginning/ending, so we remove these blanks from start and end...
	return txt_new.stripWhiteSpace();


}
//---------------------------------------------------------------------------


void ProtoTypeDoc::setConfigMaxCharactersPerLine(int cpl)
{
	config_max_characters_per_line = cpl;

}
//---------------------------------------------------------------------------


Block* ProtoTypeDoc::getBlock( int id )
{

	Block* block = NULL;

	int b = blocklist->find( new Block( id ) );
	if ( b > -1 ){
		block = blocklist->at( b );
	}

	return block;

}
//---------------------------------------------------------------------------



/** No descriptions */
QString ProtoTypeDoc::getCurrentProtocolFilename()
{
  return currentProtocolFilename;
}
